# -*- coding: utf-8 -*-

# Copyright 2011 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import sys
from skoolparser import DIRECTIVES, get_address, parse_int

VALID_CTLS = DIRECTIVES + ' *'
VERBATIM_BLOCKS = 'dir'

def get_defb_length(op):
    return op.count(',') + 1

def get_defw_length(op):
    return 2 * (op.count(',') + 1)

def get_defm_length(op):
    start = op.index('"')
    return op.index('"', start + 1) - start - 1

class Line:
    def is_ctl_line(self):
        return False

class VerbatimLine(Line):
    def __init__(self, text):
        self.text = text

    def __str__(self):
        return self.text

class ControlLine(Line):
    def __init__(self, ctl, address, addr_str, operation, comment_index, comment):
        self.ctl = ctl
        self.address = address
        self.addr_str = addr_str
        self.comment_index = comment_index
        self.comment = comment
        self.operation = operation
        self.inst_ctl = ' ' if ctl == 'i' else self._get_instruction_ctl()
        self.lengths = None

    def __str__(self):
        comment = (' %s' % self.comment).rstrip()
        comment_index = ';%i' % self.comment_index if self.comment_index > 0 else ''
        if self.lengths:
            return "%s%s%s,%s%s%s\n" % (self.ctl, self.inst_ctl, self.addr_str, self._get_lengths(), comment_index, comment)
        return "%s%s%s%s%s\n" % (self.ctl, self.inst_ctl, self.addr_str, comment_index, comment)

    def _get_lengths(self):
        # Find subsequences of indentical statement lengths and abbreviate them,
        # e.g. '16,16,16,8,8,4' -> '16*3,8*2,4'
        if self.inst_ctl == 'C':
            return sum(self.lengths)
        lengths = []
        prev = None
        for length in self.lengths:
            if length == prev:
                lengths[-1][1] += 1
            else:
                lengths.append([length, 1])
                prev = length
        length_params = []
        for length, mult in lengths:
            if mult == 1:
                length_params.append(str(length))
            else:
                length_params.append('%i*%i' % (length, mult))
        return ','.join(length_params)

    def _get_instruction_ctl(self):
        op = self.operation.upper()
        if op.startswith('DEFB'):
            return 'B'
        if op.startswith('DEFW'):
            return 'W'
        if op.startswith('DEFM'):
            return 'T'
        if op.startswith('DEFS'):
            return 'Z'
        return 'C'

    def calculate_length(self, next_ctl_line):
        if self.inst_ctl == 'C':
            if next_ctl_line:
                length = next_ctl_line.address - self.address
            else:
                length = -1
            if length < 0 or length > 4:
                length = 1
        elif self.inst_ctl == 'B':
            length = get_defb_length(self.operation)
        elif self.inst_ctl == 'T':
            length = get_defm_length(self.operation)
        elif self.inst_ctl == 'W':
            length = get_defw_length(self.operation)
        else:
            length = parse_int(get_address(self.operation[4:]))
        self.lengths = [length]

    def is_ctl_line(self):
        return True

class SftWriter:
    def __init__(self, skoolfile, write_hex=False):
        self.skoolfile = skoolfile
        self.write_hex = write_hex
        self.stack = []
        self.verbatim = False

    def _parse_skool(self):
        lines = []
        ctl_lines = []
        entry_ctl = None
        f = open(self.skoolfile, 'r')
        for line in f:
            if line.startswith(';'):
                comment = line[1:].strip()
                if comment.startswith('@'):
                    lines.append(VerbatimLine(line))
                    self._parse_asm_directive(comment[1:])
                    continue
            if self.verbatim:
                # This line is inside a '+' block, so include it as is
                lines.append(VerbatimLine(line))
                continue
            s_line = line.strip()
            if not s_line:
                # This line is blank
                lines.append(VerbatimLine(line))
                entry_ctl = None
                continue
            # Check whether we're in a block that should be preserved verbatim
            if entry_ctl is None and line[0] in VERBATIM_BLOCKS:
                entry_ctl = line[0]
            if entry_ctl and entry_ctl in VERBATIM_BLOCKS:
                lines.append(VerbatimLine(line))
                continue
            if s_line[0] == ';':
                if line[0] == ';':
                    # This line is a comment
                    lines.append(VerbatimLine(line))
                else:
                    # This line is a continuation of an instruction comment
                    comment_index = line.index(';')
                    lines.append(VerbatimLine(" ;%i %s" % (comment_index, line[comment_index + 1:].lstrip())))
                continue
            # Check whether the line starts with a valid character
            if line[0] not in VALID_CTLS:
                lines.append(VerbatimLine(line))
                continue
            # This line contains an instruction
            ctl_line = self._parse_instruction(line)
            lines.append(ctl_line)
            ctl_lines.append(ctl_line)
        f.close()
        self._calculate_lengths(ctl_lines)
        return self._compress_blocks(lines)

    def _parse_instruction(self, line):
        ctl = line[0]
        address = parse_int(line[1:6])
        addr_str = self._addr_str(address)
        comment_index = line.find(';', 7)
        end = comment_index if comment_index > 0 else len(line)
        operation = line[7:end].strip()
        comment = line[end + 1:].strip()
        return ControlLine(ctl, address, addr_str, operation, comment_index, comment)

    def _parse_asm_directive(self, directive):
        prefix = directive[:4]
        infix = directive[len(prefix):len(prefix) + 1]
        suffix = directive[len(prefix) + len(infix):].rstrip()
        if prefix in ('ofix', 'bfix', 'rfix', 'isub', 'rsub') and infix in '+-' and suffix in ('begin', 'else', 'end'):
            cur_op = self.stack[-1] if self.stack else (None, None)
            if suffix == 'begin':
                if prefix == cur_op[0]:
                    _error('%s inside %s%s block' % (directive, cur_op[0], cur_op[1]))
                self.stack.append((prefix, infix))
            elif suffix == 'else':
                if cur_op[0] is None:
                    _error('%s not inside block' % directive)
                if prefix != cur_op[0] or infix == cur_op[1]:
                    _error('%s inside %s%s block' % (directive, cur_op[0], cur_op[1]))
                self.stack[-1] = (prefix, infix)
            elif suffix == 'end':
                if cur_op[0] is None:
                    _error('%s has no matching start directive' % directive)
                if (prefix, infix) != cur_op:
                    _error('%s cannot end %s block' % (directive, cur_op))
                self.stack.pop()

            include = True
            for p, i in self.stack:
                include = i == '-'
                if not include:
                    break

            self.verbatim = not include

    def _calculate_lengths(self, ctl_lines):
        for i, ctl_line in enumerate(ctl_lines):
            if ctl_line.ctl != 'i':
                next_ctl_line = ctl_lines[i + 1] if i < len(ctl_lines) - 1 else None
                ctl_line.calculate_length(next_ctl_line)

    def _compress_blocks(self, lines):
        """Compress sequences of commentless lines into a single line."""
        compressed = []
        prev_line = None
        for line in lines:
            if line.is_ctl_line() and not line.comment and (line.ctl == ' ' or line.ctl in DIRECTIVES):
                if prev_line is None:
                    prev_line = line
                elif prev_line.inst_ctl == line.inst_ctl and prev_line.comment_index == line.comment_index and line.address == prev_line.address + sum(prev_line.lengths):
                    prev_line.lengths.append(line.lengths[0])
                else:
                    compressed.append(prev_line)
                    prev_line = line
            else:
                if prev_line:
                    compressed.append(prev_line)
                compressed.append(line)
                prev_line = None
        if prev_line:
            compressed.append(prev_line)
        return compressed

    def _addr_str(self, address):
        if address is not None:
            return '$%04X' % int(address) if self.write_hex else str(address)

    def write(self, out=None):
        if out is None:
            out = sys.stdout
        out.writelines(str(line) for line in self._parse_skool())
