#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2009-2011 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import sys
from os.path import isfile, basename
from skoolkit.snaskool import SkoolWriter, SftParser, Recipe, get_snapshot, parse_ctl, generate_ctl

START = 16384
DEFB_SIZE = 8
DEFM_SIZE = 66

class Options:
    def __init__(self):
        self.ctlfile = None
        self.sftfile = None
        self.genctl = False
        self.genctlfile = None
        self.ctl_hex = False
        self.asm_hex = False
        self.asm_lower = False
        self.start = START
        self.org = None
        self.text = False
        self.write_refs = True
        self.defb_size = DEFB_SIZE
        self.defb_mod = 1
        self.defm_width = DEFM_SIZE
        self.zfill = False

def find(fname):
    if isfile(fname):
        return fname

def parse_args(args):
    options = Options()
    files = []
    i = 0
    while i < len(args):
        arg = args[i]
        if arg == '-c':
            options.ctlfile = args[i + 1]
            if options.ctlfile != '-' and not isfile(options.ctlfile):
                sys.stderr.write('ERROR: %s: file not found\n' % options.ctlfile)
                sys.exit(1)
            i += 1
        elif arg == '-T':
            options.sftfile = args[i + 1]
            if options.sftfile != '-' and not isfile(options.sftfile):
                sys.stderr.write('ERROR: %s: file not found\n' % options.sftfile)
                sys.exit(1)
            i += 1
        elif arg == '-g':
            options.genctl = True
            options.genctlfile = args[i + 1]
            i += 1
        elif arg == '-h':
            options.ctl_hex = True
        elif arg == '-H':
            options.asm_hex = True
        elif arg == '-L':
            options.asm_lower = True
        elif arg == '-s':
            options.start = int(args[i + 1])
            i += 1
        elif arg == '-o':
            options.org = int(args[i + 1])
            i += 1
        elif arg == '-t':
            options.text = True
        elif arg == '-r':
            options.write_refs = False
        elif arg == '-n':
            options.defb_size = int(args[i + 1])
            i += 1
        elif arg == '-m':
            options.defb_mod = int(args[i + 1])
            i += 1
        elif arg == '-l':
            options.defm_width = int(args[i + 1])
            i += 1
        elif arg == '-z':
            options.zfill = True
        elif arg[0] == '-':
            print_usage()
        else:
            files.append(arg)
        i += 1

    if len(files) != 1 or files[0][-4:] not in ('.bin', '.sna', '.z80'):
        print_usage()
    snafile = files[0]
    prefix = snafile[:-4]
    if options.ctlfile is None:
        options.ctlfile = find('%s.ctl' % prefix)
    if options.sftfile is None:
        options.sftfile = find('%s.sft' % prefix)
    return snafile, options

def print_usage():
    sys.stderr.write("""Usage: %s [options] file

  Convert a binary (raw memory) file or a 48K SNA or Z80 snapshot into a skool
  file.

Options:
  -c FILE  Use FILE as the control file (default is file.ctl)
  -T FILE  Use FILE as the skool file template (default is file.sft)
  -g FILE  Generate a control file in FILE
  -h       Write hexadecimal addresses in the generated control file
  -H       Write hexadecimal addresses and operands in the disassembly
  -L       Write the disassembly in lower case
  -s ADDR  Specify the address at which to start disassembling (default=%i)
  -o ADDR  Specify the origin address of file.bin (default: 65536 - length)
  -t       Show ASCII text in the comment fields
  -r       Suppress addition of comments listing entry point referrers
  -n N     Set the max number of bytes per DEFB statement to N (default=%i)
  -m M     Group DEFB blocks by addresses that divide M
  -z       Write bytes with leading zeroes in DEFB statements
  -l L     Set the max number of characters per DEFM statement to L (default=%i)
"""  % (basename(sys.argv[0]), START, DEFB_SIZE, DEFM_SIZE))
    sys.exit(1)

def run(snafile, options):
    # Read the snapshot file
    if snafile[-4:] == '.bin':
        f = open(snafile, 'rb')
        ram = f.read()
        f.close()
        org = 65536 - len(ram) if options.org is None else options.org
        snapshot = [0] * org + [ord(b) for b in ram]
        end = len(snapshot)
        snapshot += [0] * (65536 - end)
        ctls = {max(org, options.start): 'c'}
        if end < 65536:
            ctls[end] = 'i'
        recipe = Recipe(ctls)
    else:
        snapshot = get_snapshot(snafile)
        recipe = Recipe({options.start: 'c'})

    # Pad out the end of the snapshot to avoid errors when an instruction
    # crosses the 64K boundary
    snapshot.extend((0, 0, 0))

    if options.sftfile:
        # Use a skool file template
        writer = SftParser(snapshot, options.sftfile, options.zfill, options.asm_hex, options.asm_lower)
        writer.write_skool()
        sys.exit()

    if options.genctl:
        # Analyse the snapshot and generate a control file
        recipe = Recipe(generate_ctl(snapshot, options.genctlfile, options.ctl_hex, options.start))
    elif options.ctlfile:
        # Use a control file
        recipe = parse_ctl(options.ctlfile)
    writer = SkoolWriter(snapshot, recipe, options.defb_size, options.defb_mod, options.zfill, options.defm_width, options.asm_hex, options.asm_lower)
    writer.write_skool(options.write_refs, options.text)

def main():
    run(*parse_args(sys.argv[1:]))

if __name__ == '__main__':
    main()
